#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <X11/Xlib.h>
#include <X11/extensions/xf86dga.h>
#include <gdk/gdktypes.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <gtk/gtkeventbox.h>

#include <linux/types.h>
#include <linux/videodev.h>
#include "channels.h"
#define min(x,y)	((x) < (y) ? (x) : (y))
#define max(x,y)	((x) > (y) ? (x) : (y))

static GtkWidget *window;
static GtkWidget *tv;
static GtkWidget *painter;
static GdkGC *painter_gc;
static GdkColor chroma_color;
static GdkImage *capture;		/* Capture Image */
static GdkImageType capture_type;
static gint capture_h,capture_w;
static gint window_h, window_w;

static int widescreen=0;
static int pseudocolor=0;
static int widescreen_h;

/*
 *	TV card driving.
 */

static struct video_window vwin;
static struct video_tuner vt;
static struct video_picture vpic;
static struct video_buffer vbuf;
static struct video_capability vcap;
static struct video_channel chan;
static struct video_audio vaudio;
#ifdef EXPERIMENT
static struct video_clip clips[256];	/* didnt want on stack */
#endif
static int grabber_format;
static int capture_running;
static int capture_hidden;	/* Dont capture we are hidden */

static int tv_fd = -1;

/*
 *	Methods to work with
 */
 
static int kill_on_overlap = 0;	/* Set for overlay cards that dont clip */
static int clean_display = 0;		/* Set on overlay cards that do DMA */
static int chroma_key = 0;		/* Set for chromakey cards */
static int fixed_size = 0;		/* Size is fixed by the card (we dont scale yet) */
static int use_shm = 0;		/* Capture only card */
static int xsize = 640, ysize = 480;	/* Initial size */
static int xmin, ymin;			/* Limits */
static int xmax, ymax;	

static int colormap_size;
static int colormap_base;
static GdkColormap *private_colormap;

static int sanity1=0;

static char buf[128];		/* temp. buffer */
static int current_ch = 0;	/* initialized by first chan up */
static int ch_type=NTSC_CABLE;	/* set via menu */
static void chan_to_bttv(int freq)
{
	int i = ch_type;
	ioctl(tv_fd, VIDIOCGTUNER, &vt);
	if (ch_type <= NTSC_JP_BCAST) {
		vt.flags=VIDEO_TUNER_NTSC;
		vt.mode=VIDEO_MODE_NTSC;
	} else {
		if (ch_type == 10) {	/* SECAM */
			vt.mode=VIDEO_MODE_SECAM;
			vt.flags=VIDEO_TUNER_SECAM;
			/* I'm not sure of tuning for SECAM */
			/* SECAM is not listed in my charts */
			i=PAL_UHF_GHI;
		} else {
			vt.mode=VIDEO_MODE_PAL;
			vt.tuner=VIDEO_TUNER_PAL;
		}
	}
	ioctl(tv_fd, VIDIOCSTUNER, &vt);
	sprintf(buf, "Miguelivison 1.04: %s Chan %s", vcap.name,
		tvtuner[freq].name);
	gtk_window_set_title(GTK_WINDOW(window), buf);
	freq = tvtuner[freq].freq[i];
	if (!freq)	/* no need to set invalid freq */
	    return;
	freq += 6250;	/* driver tunings all seem offset */
	freq *= 16;
	freq /= 1000;
	/* TODO: add single-step fine-tune -- need slider */
	ioctl(tv_fd, VIDIOCSFREQ, &freq);
}
static void channel_up(void)
{
	if ((++current_ch) == CHAN_ENTRIES)
		current_ch = 0;
	while (tvtuner[current_ch].freq[ch_type] == 0) {
		current_ch++;
		if (current_ch == CHAN_ENTRIES)
			current_ch = 0;
	}
	chan_to_bttv(current_ch);
}
static void channel_down(void)
{
	if ((--current_ch) < 0)
		current_ch = CHAN_ENTRIES-1;
	while (tvtuner[current_ch].freq[ch_type] == 0) {
		current_ch--;
		if (current_ch < 0)
			current_ch = CHAN_ENTRIES-1;
	}
	chan_to_bttv(current_ch);
}

static void set_format(GtkObject *o, void *p)
{
	
	ch_type = (int)p;
	chan_to_bttv(current_ch);
}


static void make_palette_grey(int num)
{
	int i,x;
	int step=256/num;
	
	private_colormap = gdk_colormap_new(gdk_visual_get_best(),num);
	for(i=0;i<num; i++)
	{
		x = i * step;
		sprintf(buf,"#%02X%02X%02X", x,x,x);
		gdk_color_parse(buf, &private_colormap->colors[i]);
		gdk_color_alloc(gdk_colormap_get_system(),
			&private_colormap->colors[i]);
	}
	colormap_size=num;
	colormap_base=0;
	gdk_colormap_change(private_colormap, 256);
	gdk_window_set_colormap(tv->window, private_colormap);
	gdk_window_set_colormap(window->window, private_colormap);
}

static void make_color_cube(void)
{
	int i,r,g,b;
	
	private_colormap = gdk_colormap_new(gdk_visual_get_best(),256);
	for(i=0;i<16;i++)
	{
		gdk_color_black(private_colormap,
			&(private_colormap->colors[i]));
	}
	for(i=0;i<216;i++)
	{
		r=i/36; g=i/6; b=i%6;
		sprintf(buf,"#%02X%02X%02X", r * 40, b * 40, g * 40);
		gdk_color_parse(buf, &private_colormap->colors[i+16]);
		gdk_color_alloc(gdk_colormap_get_system(),
			&private_colormap->colors[i]);
	}
	colormap_size=216;
	colormap_base=16;
	gdk_colormap_change(private_colormap, 256);
	gdk_window_set_colormap(tv->window, private_colormap);
	gdk_window_set_colormap(window->window, private_colormap);
}

static char *probe_tv_set(int n)
{
	static struct video_capability t;
	int fd;
	sprintf(buf,"/dev/bttv%d", n);
	
	fd=open(buf, O_RDWR);
	if(fd==-1)
		return NULL;
		
	if(ioctl(fd, VIDIOCGCAP, &t)==-1)
	{
		close(fd);
		return NULL;
	}
	close(fd);
	return t.name;
}

static int open_tv_card(int n)
{
	Display *disp;
	int fd;
	unsigned int rwidth;
	int bank, ram, major, minor;
	
	
	sprintf(buf,"/dev/bttv%d",n);
	
	fd=open(buf,O_RDWR);
	
	
	if(fd==-1)
	{
		perror("open");
		return -1;
	}
	
	if(ioctl(fd, VIDIOCGAUDIO, &vaudio)==0)
	{
		vaudio.flags|=VIDEO_AUDIO_MUTE;
		ioctl(fd, VIDIOCSAUDIO, &vaudio);
	}

	if(ioctl(fd, VIDIOCGCAP, &vcap))
	{
		perror("get capabilities");
		return -1;
	}
	
	/*
	 *	Use black as the background unless we later find a 
	 *	chromakey device
	 */
	 
	gdk_color_black(gdk_colormap_get_system(), &chroma_color);
	 
	/* 
	 *	Now see what we support
	 */
	 
	if(!(vcap.type&VID_TYPE_OVERLAY))
	{
		if(vcap.type&VID_TYPE_CAPTURE)
			use_shm = 1;
		else
		{
			fprintf(stderr,"I don't know how to handle this device.\n");
			return -1;
		}
	}
	else
	{
		/* Ok now see what method to use */
		if(vcap.type&VID_TYPE_CLIPPING)
		{
		}
		else if(vcap.type&VID_TYPE_CHROMAKEY)
		{
			chroma_key = 1;
			/* We want the most vile unused color we can get
			   for our chroma key - this is pretty foul */
			gdk_color_parse("#FA77A1", &chroma_color);
			gdk_color_alloc(gdk_colormap_get_system(), &chroma_color);
		}
		else
			kill_on_overlap = 1;
		if(vcap.type&VID_TYPE_FRAMERAM)
			clean_display = 1;
		
	}
	
	/*
	 *	Set the window size constraints
	 */
	 
	if(!(vcap.type&VID_TYPE_SCALES))
	{
		xmin = xmax = xsize = vcap.maxwidth;
		ymin = ymax = ysize = vcap.maxheight;
		fixed_size = 1;
	}
	else
	{
		xmin = vcap.minwidth;
		ymin = vcap.minheight;
		xmax = vcap.maxwidth;
		ymax = vcap.maxheight;
		xsize = min(max(xsize, xmin), xmax);
		ysize = min(max(ysize, ymin), ymax);
	}

	/*
	 *	Now get the video parameters (one day X will set these)
	 */
	 
	disp=GDK_DISPLAY();
	if (XF86DGAQueryVersion(disp, &major, &minor)) 
		XF86DGAGetVideoLL(disp, DefaultScreen(disp),
		(int *) &(vbuf.base), &rwidth, &bank, &ram);
	else 
	{
		fprintf(stderr,"XF86DGA: DGA not available.\n");
		return -1;
	}

	/*
	 *	Hunt the visual
	 */
	 
	vbuf.depth = gdk_visual_get_best_depth();
	vbuf.bytesperline = rwidth*vbuf.depth/8;
	
	/*
	 *	Set the overlay buffer up
	 */
	 
	if(ioctl(fd, VIDIOCSFBUF, &vbuf)==-1)
	{
		perror("Set frame buffer");
		/*
		 *	If we can't use MITSHM then we can't overlay if
		 *	the board can't handle our framebuffer...
		 *
		 *	For SHM we can do post processing.
		 */
		 
		if(!use_shm)
			return -1;
	}
	
	if(seteuid(getuid())==-1)
	{
		perror("seteuid");
		exit(1);
	}
	
	/*
	 *	Set the format
	 */
 
	ioctl(fd, VIDIOCGPICT , &vpic);

	if(use_shm)
	{	
	
		/*
		 *	Try 24bit RGB, then 565 then 555. Arguably we ought to
		 *	try one matching our visual first, then try them in 
		 *	quality order
		 *
		 */
		
		
		if(vcap.type&VID_TYPE_MONOCHROME)
		{
			vpic.depth=8;
			vpic.palette=VIDEO_PALETTE_GREY;	/* 8bit grey */
			if(ioctl(fd, VIDIOCSPICT, &vpic)==-1)
			{
				vpic.depth=6;
				if(ioctl(fd, VIDIOCSPICT, &vpic)==-1)
				{
					vpic.depth=4;
					if(ioctl(fd, VIDIOCSPICT, &vpic)==-1)
					{
						fprintf(stderr, "Unable to find a supported capture format.\n");
						return -1;
					}
				}
			}
			/*
			 *	Make a grey scale cube - only go to 6bit
			 */
			 
			if(vpic.depth==4)
				make_palette_grey(16);
			else
				make_palette_grey(64);
			
		}
		else
		{
			vpic.depth=24;
			vpic.palette=VIDEO_PALETTE_RGB24;
		
			if(ioctl(fd, VIDIOCSPICT, &vpic)==-1)
			{
				vpic.palette=VIDEO_PALETTE_RGB565;
				vpic.depth=16;
				
				if(ioctl(fd, VIDIOCSPICT, &vpic)==-1)
				{
					vpic.palette=VIDEO_PALETTE_RGB555;
					vpic.depth=15;
				
					if(ioctl(tv_fd, VIDIOCSPICT, &vpic)==-1)
					{
						fprintf(stderr, "Unable to find a supported capture format.\n");
						return -1;
					}
				}
			}
		}
	}
	/*
	 *	Soft overlays by chroma etc we dont have to care about
	 *	the format for.. but DMA to frame buffer we do
	 */
	else if(vcap.type&VID_TYPE_FRAMERAM)
	{
		vpic.depth = vbuf.depth;
		switch(vpic.depth)
		{
			case 8:
				vpic.palette=VIDEO_PALETTE_HI240;
				pseudocolor++;
				break;
			case 15:
				vpic.palette=VIDEO_PALETTE_RGB555;
				break;
			case 16:
				vpic.palette=VIDEO_PALETTE_RGB565;
				break;
			case 24:
				vpic.palette=VIDEO_PALETTE_RGB24;
				break;
			case 32:		
				vpic.palette=VIDEO_PALETTE_RGB32;
				break;
			default:
				fprintf(stderr,"Unsupported X11 depth.\n");
		}
		if(ioctl(fd, VIDIOCSPICT, &vpic)==-1)
		{
			/* Depth mismatch is fatal on overlay */
			perror("set depth");
			return -1;
		}
	}
		
	if(ioctl(fd, VIDIOCGPICT, &vpic)==-1)
		perror("get picture");
			
	grabber_format = vpic.palette;
	
	sanity1=1;
	return fd;
}	


static void close_tv_card(int handle)
{
	close(handle);
}

/*
 *	Capture an image from capture cards - idea and X11 side taken
 *	from the excellent example in sane-0.67/xcam
 */

#define READ_VIDEO_PIXEL(buf, format, depth, r, g, b)			\
{									\
	switch (format)							\
	{								\
		case VIDEO_PALETTE_GREY:				\
			switch (depth)					\
			{						\
				case 4:					\
				case 6:					\
				case 8:					\
					(r) = (g) = (b) = (*buf++ << 8);\
					break;				\
									\
				case 16:				\
					(r) = (g) = (b) = 		\
						*((guint16 *) buf);	\
					buf += 2;			\
					break;				\
			}						\
			break;						\
									\
									\
		case VIDEO_PALETTE_RGB565:				\
		{							\
			unsigned short tmp = *(unsigned short *)buf;	\
			(r) = tmp&0xF800;				\
			(g) = (tmp<<5)&0xFC00;				\
			(b) = (tmp<<11)&0xF800;				\
			buf += 2;					\
		}							\
		break;							\
									\
		case VIDEO_PALETTE_RGB555:				\
			(r) = (buf[0]&0xF8)<<8;				\
			(g) = ((buf[0] << 5 | buf[1] >> 3)&0xF8)<<8;	\
			(b) = ((buf[1] << 2 ) & 0xF8)<<8;		\
			buf += 2;					\
			break;						\
									\
		case VIDEO_PALETTE_RGB24:				\
			(r) = buf[0] << 8; (g) = buf[1] << 8; 		\
			(b) = buf[2] << 8;				\
			buf += 3;					\
			break;						\
									\
		default:						\
			fprintf(stderr, 				\
				"Format %d not yet supported\n",	\
				format);				\
	}								\
}						


#define PUT_X11_PIXEL(buf, endian, depth, bpp, r, g, b, gl_map)		\
{									\
	switch (depth)							\
	{								\
		case  1: /* duh?  A Sun3 or what?? */			\
			lum = 3*(r) + 5*(g) + 2*(b);			\
			if (lum >= 5*0x8000)				\
				*buf |= dst_mask;			\
			dst_mask <<= 1;					\
			if (dst_mask > 0xff)				\
			{						\
				buf += (bpp);				\
				dst_mask = 0x01;			\
			}						\
			break;						\
									\
		case  8:	/* FIXME: Mono only right now */	\
			lum = ((3*(r) + 5*(g) + 2*(b)) / (10 * 256/colormap_size)) >> 8;	\
			if (lum >= colormap_size)			\
				lum = colormap_size;			\
			buf[0] = private_colormap->colors[lum+colormap_base].pixel;	\
			buf += (bpp);					\
			break;						\
		case 15:						\
			rgb = (  (((r) >> 11) << 10)	/* 5 bits of red */		\
			| (((g) >> 11) <<  5)	/* 5 bits of green */	\
			| (((b) >> 11) <<  0));	/* 5 bits of blue */	\
			((guint16 *)buf)[0] = rgb;			\
			buf += (bpp);					\
			break;						\
									\
		case 16:						\
			rgb = (  (((r) >> 11) << 11)	/* 5 bits of red */	\
			| (((g) >> 10) <<  5)	/* 6 bits of green */	\
			| (((b) >> 11) <<  0));	/* 5 bits of blue */	\
			((guint16 *)buf)[0] = rgb;			\
			buf += (bpp);					\
			break;						\
									\
		case 24:						\
			/* Is this correctly handling all byte order cases? */		\
			if ((endian) == GDK_LSB_FIRST)			\
			{						\
				buf[0] = (b) >> 8; buf[1] = (g) >> 8; buf[2] = (r) >> 8;\
			}						\
			else						\
			{						\
				buf[0] = (r) >> 8; 			\
				buf[1] = (g) >> 8; 			\
				buf[2] = (b) >> 8;			\
			}						\
			buf += (bpp);					\
			break;						\
									\
		case 32:						\
      /* Is this correctly handling all byte order cases?  It assumes	\
	 the byte order of the host is the same as that of the		\
	 pixmap. */							\
			rgb = (((r) >> 8) << 16) | (((g) >> 8) << 8) | ((b) >> 8);\
			((guint32 *)buf)[0] = rgb;			\
			buf += (bpp);					\
			break;						\
		default:						\
			fprintf(stderr, "Unsupported X11 depth.\n");	\
	}								\
}
 
static void grab_image(unsigned char *dst, int xpos, int ypos)
{
	int x;
	int lines=0;
	int pixels_per_line, bytes_per_line, byte_order, bytes_per_pixel;
	int dst_depth, src_depth;
	guint32 r = 0, g = 0, b = 0, lum, rgb, src_mask, dst_mask;
	unsigned char *mem, *src;	

	src_depth = vpic.depth;
	dst_depth = capture->depth;
	
	pixels_per_line = capture_w;
	bytes_per_line = capture->bpl;
	bytes_per_pixel = capture->bpp;
	byte_order = capture->byte_order;

	mem=malloc(bytes_per_line * capture_h);
	if(mem==NULL)
		return;

	src=mem;
	
	read(tv_fd, mem, bytes_per_line * capture_h);
			
	x = 0;

	src_mask = 0x80;
	dst_mask = 0x01;
	
	while (lines<capture_h)
	{
		READ_VIDEO_PIXEL(src, grabber_format, src_depth, r, g, b);
		PUT_X11_PIXEL(dst, byte_order, dst_depth, bytes_per_pixel, r, g, b,
			win.canvas.graylevel);
		if (++x >= pixels_per_line)
		{
			x = 0;
			dst += bytes_per_line - pixels_per_line * bytes_per_pixel;
			lines++;
		}
	}
	free(mem);
}

static void write_ppm_image(FILE *f)
{
	int x;
	int lines=0;
	int pixels_per_line, bytes_per_line, byte_order, bytes_per_pixel;
	int dst_depth, src_depth;
	guint32 r = 0, g = 0, b = 0, src_mask, dst_mask;
	unsigned char *mem, *src;	

	src_depth = vpic.depth;
	dst_depth = capture->depth;
	
	pixels_per_line = capture_w;
	bytes_per_line = capture->bpl;
	bytes_per_pixel = capture->bpp;
	byte_order = capture->byte_order;

	mem=malloc(bytes_per_line * capture_h);
	if(mem==NULL)
		return;

	src=mem;
	
	read(tv_fd, mem, bytes_per_line * capture_h);
			
	x = 0;

	src_mask = 0x80;
	dst_mask = 0x01;
	
	fprintf(f,"P6 %d %d 255\n",
		pixels_per_line, capture_h);
	
	while (lines<capture_h)
	{
		READ_VIDEO_PIXEL(src, grabber_format, src_depth, r, g, b);
		fputc(r>>8, f);
		fputc(g>>8, f);
		fputc(b>>8, f);
		if (++x >= pixels_per_line)
		{
			x=0;
			lines++;
		}
	}
	free(mem);
	fclose(f);
}


/*
 *	Paint the drawing area. We use this to keep it clean when not
 *	capturing, for chromakey areas and for mitshm.
 */
 
static void do_painting(void)
{
	gint x,y,w,h,d;
	
	/*
	 *	Called during setup - ignore	 
	 */
	 
	if(painter_gc==NULL)
		return;
		
		
        gdk_window_get_geometry(painter->window, &x, &y, &w, &h, &d);

	if(use_shm && capture && capture_running && !capture_hidden)
	{
		if(widescreen)
		{
			gdk_draw_image(painter->window, painter_gc, capture,
				0, capture_h/10, 
				0, 0,
				capture_w,
				widescreen_h);
		}
		else
		{
			gdk_draw_image(painter->window, painter_gc, capture,
				0,0,0,0, 
				capture_w, capture_h);
		}
	}
	else
		gdk_draw_rectangle(painter->window, painter_gc, 1, 0, 0, w, h);
	gdk_flush();
}
		

static void tv_grabber(void)
{
	grab_image(capture->mem, 0, 0);
	do_painting();
}

static void tv_capture(int handle, int onoff)
{
#ifdef EXPERIMENT
	int old_state = onoff;
#endif
	
	if(use_shm==0)
	{
		/*
		 *	Dont capture if unmapped
		 */

		capture_running = onoff;

		if(capture_hidden)
			onoff=0;
			
#ifdef EXPERIMENT			
		if (onoff!=old_state)
		{
			if(onoff == 1)
			{
				int zero=0;
				ioctl(handle, VIDIOCCAPTURE, &zero);
				gdk_window_set_override_redirect(painter->window, TRUE);
				gdk_window_reparent(painter->window, GDK_ROOT_PARENT(),vwin.x, vwin.y); 
				gdk_window_raise(painter->window);
			}
			else
			{
				gdk_window_set_override_redirect(painter->window, FALSE);
				gdk_window_reparent(painter->window, painter->parent->window, painter->allocation.x, painter->allocation.y); 
				gdk_window_raise(painter->window);
			}
		}
#endif				
		if(ioctl(handle, VIDIOCCAPTURE, &onoff)==-1)
			perror("Capture");
		printf("Capture is now %d\n",  onoff);
	}
	else
	{
		static gint capture_tag;
		if(capture_running && capture && capture->mem)
			grab_image(capture->mem, 0,0);
		if(onoff)
			capture_tag = gtk_idle_add((GtkFunction)tv_grabber, 0);
		else
			gtk_idle_remove(capture_tag);
		capture_running = onoff;
	}
}


static void tv_set_window(int handle)
{
	gint x,y,w,h,d;
	GdkWindow *t=painter->window;
	
	/*
	 *	GDK holds the window positioning
	 */
	
        gdk_window_get_geometry( t, &x, &y, &w, &h, &d);
        
        	
        /*
         *	We are working on screen positions with the overlay
         *	devices, so see where we are relative to the screen.
         */
        
        gdk_window_get_origin(t, &x, &y);

#ifdef DEBUG
        printf("Window is %d by %d at %d,%d\n", w,h,x,y);
#endif
	
	if(w < xmin || h < ymin)
	{
		/* Too small, turn it off */
		tv_capture(handle,0);
	}
	
	/* Constrain the window */
	
	h=min(h,ymax);
	w=min(w,xmax);
	
	window_h=h;
	window_w=w;
	widescreen_h=h;
	
	vwin.x = x;
	vwin.y = y;
	vwin.width = w;
	vwin.height = h;
	vwin.clips = NULL; 
	vwin.flags = 0;
	vwin.clipcount = 0;
	
	if(h<32 || w<32)
		return;

	if(widescreen)
	{
		/*
		 *	In wide screen mode we want to overcapture vertically
		 *	by about 10% above and below.
		 */
		vwin.y-=h/10;
		vwin.height+=h/5;
	}
	if(sanity1!=0)
	{
		if(ioctl(tv_fd, VIDIOCSWIN, &vwin)==-1)
			perror("set video window");
		if(ioctl(tv_fd, VIDIOCGWIN, &vwin)==-1)
			perror("get video window");
		/* x,y,h,w are now the values that were chosen by the
		   device to be as close as possible - eg the quickcam only
		   has 3 sizes */
		xsize = vwin.width;
		ysize = vwin.height;

		/*
		 *	Now check the image in question
		 */
		 
		if(use_shm)
		{
			if(capture==NULL || capture_w!=xsize || capture_h!=ysize)
			{
				capture_type = GDK_IMAGE_FASTEST;
				
				if(capture!=NULL)
				{
					gdk_image_destroy(capture);
					capture=NULL;
				}
#ifdef __alpha__
				/*
				 *	Taken from the sane xcam
				 */
				 
				/* Some X servers seem to have a problem with shared images that
				    have a width that is not a multiple of 8.  Duh... ;-( */
				if (xsize % 8)
					capture_type = GDK_IMAGE_NORMAL;
#endif
				
				capture=gdk_image_new(capture_type, 
					gdk_window_get_visual(t),
					xsize,ysize);
					
				capture_h=ysize;
				capture_w=xsize;
			}
			/*
			 *	Capture and draw
			 */
			 
			d=capture_running;
			capture_running=0;
			do_painting();
			capture_running=d;
			
			/*
			 *	Start grabbing again
			 */
			
			grab_image(capture->mem, x,y);
			
		}				
	}
}

static int has_shrunk(GtkWidget *wp)
{
	int w,h;
	
	gdk_window_get_size(wp->window, &w,&h);
	
	if(w<window_w || h<window_h)
		return 1;
	return 0;
}

static void show_pos(void)
{
	tv_set_window(tv_fd);
}


static void capture_on(void)
{
	tv_set_window(tv_fd);
	tv_capture(tv_fd, 1);
	do_painting();
}

static void capture_off(void)
{
	tv_capture(tv_fd,0);
	do_painting();
}

static void set_audio(GtkObject *foo, void *bar)
{
	static int audio_on=0;
	int n=(int)bar;
	vaudio.audio = 0;	/* For now FIXME FIXME */
	if(ioctl(tv_fd, VIDIOCGAUDIO, &vaudio)==-1)
	{
		perror("get audio");
		return;
	}
	
	if(n==-1)
		n = 1-audio_on;
	if(n)
		vaudio.flags&=~VIDEO_AUDIO_MUTE;
	else
		vaudio.flags|=VIDEO_AUDIO_MUTE;
		
	audio_on = n;
		
	if(ioctl(tv_fd, VIDIOCSAUDIO, &vaudio)==-1)
		perror("set audio");
}
		
static void configure_event(void)
{
#ifdef DEBUG
	printf("Got a configure event");
#endif
	tv_set_window(tv_fd);
}


/*
 *	Check for a repaint job
 */
  
static gint painting_event(GtkWidget *self, GdkEvent *event)
{
#ifdef EXPERIMENT
	GList *winlist;
	GList *node;
	GdkWindow *win;
	int next_clip = 0;
#endif
	/* Remember if we are drawing */
	int capture_was_running = capture_running;
	int crapped_on = 0;
	
#ifdef DEBUG
	printf("event %d\n", event->type);
#endif
	
	switch(event->type)
	{
		case GDK_VISIBILITY_NOTIFY:
		
			/* Boring visibility event */
			if(self != painter)
				return FALSE;

			/*
			 *	It might look like shared memory cards
			 *	dont need handling at all here, but we want
			 *	 to turn CPU draining captures off whenever
			 *	we can.
			 */
			 				
			switch(event->visibility.state)
			{
				case GDK_VISIBILITY_UNOBSCURED:
					printf("Window now unobscured\n");
					capture_hidden=0;
					tv_capture(tv_fd, capture_running);
					break;
				case GDK_VISIBILITY_PARTIAL:
					if(use_shm)
					{
						capture_hidden=0;
						tv_capture(tv_fd, capture_running);
						break;
					}
					if(vcap.type&VID_TYPE_CLIPPING)
					{
						/*
						 *	FIXME: write the code here for clipping capable cards
						 */
					}
					/* Fall through. If we cant clip then
					   we stop on partial overrides - messy but
					   unavoidable */
				case GDK_VISIBILITY_FULLY_OBSCURED:
					/*
					 *	We aren't here. Turn off capturing
					 *	to save resources and to avoid
					 *	cards peeing into video space.
					 */
					printf("Window has obscurances\n");
					if(capture_hidden)
						break;
					capture_hidden=1;
					tv_capture(tv_fd, capture_running);
					if(capture_running && clean_display)
					{
						crapped_on=clean_display;
						if(crapped_on)
							system("/usr/X11/bin/xrefresh");

					}
				default:;
			}
			break;
			
		case GDK_UNMAP:
			printf("UNMAP - turning the capture off.\n");
			capture_hidden=1;
			tv_capture(tv_fd,capture_running);
			if(capture_running && clean_display)
			{
				crapped_on = clean_display;
				if(crapped_on && capture_was_running)
				{
					printf("Refreshing\n");
					system("/usr/X11/bin/xrefresh");
				}
			}
			break;
		case GDK_MAP:
			printf("MAP - turning on the tv if needed.\n");
			capture_hidden=0;
			tv_set_window(tv_fd);
			tv_capture(tv_fd,capture_running);
			goto draw_it;
		case GDK_CONFIGURE:
			/* Fall through for now. We should only set crapped_on
			   if no longer overlap our old area */
		/*
		 *	A no expose is as good as an expose from our viewpoint
		 *	in fact it warrants more work cos we've just farted
		 *	on someone elses windows until we add the video stuff
		 *	to the X server
		 */
		 
		case GDK_NO_EXPOSE:
			printf("Got a NO EXPOSE - may be dirty.\n");
			crapped_on = clean_display;
			
		case GDK_EXPOSE:
			if(self != painter)
				return FALSE;
draw_it:
			printf("Doing EXPOSE drawing\n");
			tv_set_window(tv_fd);
			if(!painter_gc)
			{
				painter_gc=gdk_gc_new(painter->window);
				gdk_gc_set_foreground(painter_gc, &chroma_color);
				gdk_gc_set_background(painter_gc, &chroma_color);
			}	
			do_painting();
			if(crapped_on && capture_was_running && has_shrunk(painter))
			{
				printf("Refreshing\n");
				system("/usr/X11/bin/xrefresh");
			}
			break;
		default:
			break;
	}
	return FALSE;
}

		
/*
 *	Widget Stuff
 */
 
void format_bar(GtkWidget *menubar)
{
	GtkWidget* menu;
	GtkWidget* menuitem;
	int n;
	static char *name[11]= {
		"NTSC BROADCAST",
		"NTSC CABLE",
		"NTSC JP BCAST",
		"NTSC JP CABLE",
		"PAL EUROPE",
		"PAL ITALY",
		"PAL NEW ZEALAND",
		"PAL AUSTRALIA",
		"PAL UHF/IRELAND",
		"PAL CABLE (B)(G)",
		"SECAM"
	};

	menu = gtk_menu_new();
	
	menuitem = gtk_menu_item_new();
	gtk_menu_append(GTK_MENU(menu), menuitem);
	gtk_widget_show(menuitem);

	for(n=0;n<11;n++)
	{
		
		menuitem = gtk_menu_item_new_with_label(name[n]);
		gtk_menu_append(GTK_MENU(menu), menuitem);
		gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                     (GtkSignalFunc) set_format,(void *)n);
		gtk_widget_show(menuitem);
	}

	menuitem = gtk_menu_item_new_with_label("Format");
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
	gtk_container_add(GTK_CONTAINER(menubar), menuitem);
	gtk_widget_show(menuitem);
}


static void set_channel(GtkObject *o, void *p)
{
	int channel=(int)p;
	if(ioctl(tv_fd, VIDIOCSCHAN, &channel)==-1)
		perror("set channel");
	if(p==0)
		chan_to_bttv(current_ch);
}

void channel_bar(GtkWidget *menubar)
{
	GtkWidget* menu;
	GtkWidget* menuitem;
	int n;

	menu = gtk_menu_new();
	
	menuitem = gtk_menu_item_new();
	gtk_menu_append(GTK_MENU(menu), menuitem);
	gtk_widget_show(menuitem);

	for(n=0;n<vcap.channels;n++)
	{
		chan.channel=n;
		if(ioctl(tv_fd, VIDIOCGCHAN, &chan)==-1)
		{
			perror("get channel");
			continue;
		}
		menuitem = gtk_menu_item_new_with_label(chan.name);
		gtk_menu_append(GTK_MENU(menu), menuitem);
		gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                     (GtkSignalFunc) set_channel,(void *)n);
		gtk_widget_show(menuitem);
	}
	
	menuitem = gtk_menu_item_new_with_label("Input");
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
	gtk_container_add(GTK_CONTAINER(menubar), menuitem);
	gtk_widget_show(menuitem);
}

GtkWidget *tv_menu_bar(void)
{
	GtkWidget* menubar;
	GtkWidget* menu;
	GtkWidget* menuitem;

	menubar = gtk_menu_bar_new();
	menu = gtk_menu_new();
	gtk_widget_show(menubar);

	menuitem = gtk_menu_item_new();
	gtk_menu_append(GTK_MENU(menu), menuitem);
	gtk_widget_show(menuitem);

	menuitem = gtk_menu_item_new_with_label("Show");
	gtk_menu_append(GTK_MENU(menu), menuitem);
	gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                     (GtkSignalFunc) show_pos, NULL);
	gtk_widget_show(menuitem);

	menuitem = gtk_menu_item_new_with_label("Capture On");
	gtk_menu_append(GTK_MENU(menu), menuitem);
	gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                     (GtkSignalFunc) capture_on, NULL);
	gtk_widget_show(menuitem);

	menuitem = gtk_menu_item_new_with_label("Capture Off");
	gtk_menu_append(GTK_MENU(menu), menuitem);
	gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                     (GtkSignalFunc) capture_off, NULL);
	gtk_widget_show(menuitem);

	if(vcap.audios!=0)
	{
		menuitem = gtk_menu_item_new_with_label("Audio On");
		gtk_menu_append(GTK_MENU(menu), menuitem);
		gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                     (GtkSignalFunc) set_audio, (void *)1);
		gtk_widget_show(menuitem);

		menuitem = gtk_menu_item_new_with_label("Audio Off");
		gtk_menu_append(GTK_MENU(menu), menuitem);
		gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
	                     (GtkSignalFunc) set_audio, (void *)0);
		gtk_widget_show(menuitem);
	}
	
	menuitem = gtk_menu_item_new_with_label("Exit");
	gtk_menu_append(GTK_MENU(menu), menuitem);
	gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
                     (GtkSignalFunc) gtk_main_quit, NULL);
	gtk_widget_show(menuitem);

	menuitem = gtk_menu_item_new_with_label("File");
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
	gtk_container_add(GTK_CONTAINER(menubar), menuitem);
	gtk_widget_show(menuitem);

	channel_bar(menubar);
	
	format_bar(menubar);
	
	return menubar;
}


static GtkWidget *color_widget=NULL;


static gint color_settings(GtkAdjustment *adj_data, void *ptr)
{
	int n=(int)ptr;
	
	adj_data->value+=0.5;
	
	switch(n)
	{
		case 0:
			vpic.brightness=adj_data->value;
			break;
		case 1:
			vpic.contrast=adj_data->value;
			break;
		case 2:
			vpic.whiteness=adj_data->value;
			break;
		case 3:
			vpic.colour=adj_data->value;
			break;
		case 4:
			vpic.hue=adj_data->value;
			break;
	}
	if(ioctl(tv_fd, VIDIOCSPICT, &vpic)==-1)
		perror("set picture");
	return TRUE;
}

static void slider_create(char *name, GtkWidget *vbox, int id, float value)
{
	GtkWidget *scale;
	GtkObject *slider;
	
	scale=gtk_label_new(name);
	gtk_box_pack_start(GTK_BOX(vbox), scale, FALSE, FALSE, 2);
	gtk_widget_show(scale);
	
	slider = gtk_adjustment_new(value, 0.0, 65535.0, 1.0, 1.0, 0.0);
	scale = gtk_hscale_new(GTK_ADJUSTMENT(slider));
	gtk_widget_set_usize(scale,200,0);
	gtk_box_pack_start(GTK_BOX(vbox), scale, FALSE, FALSE, 0);
	
	gtk_range_set_update_policy(GTK_RANGE(scale), GTK_UPDATE_CONTINUOUS);
	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_TOP);
	gtk_scale_set_digits(GTK_SCALE(scale), 1);
	
	gtk_signal_connect(slider, "value_changed",
		(GtkSignalFunc)color_settings, (void *)id); 

	gtk_widget_show(scale);
}

static void done_color_settings(void)
{
	gtk_widget_hide(color_widget);
}


static void color_setting(void)
{
	GtkWidget *main_vbox, *vbox, *hbox, *button;
	
	
	if(color_widget)
	{
		gtk_widget_show(color_widget);
		return;
	}
	
	color_widget = gtk_dialog_new();
	
	gtk_window_set_title(GTK_WINDOW(color_widget),"Picture Control");
	main_vbox = GTK_DIALOG(color_widget)->vbox;
	
	vbox = gtk_vbox_new(FALSE,5);
	gtk_container_border_width(GTK_CONTAINER(vbox),3);
	gtk_box_pack_start(GTK_BOX(main_vbox), vbox, TRUE, TRUE, 0);
	gtk_widget_show(vbox);
	
	slider_create("Brightness", vbox, 0, (float)vpic.brightness);
	slider_create("Contrast", vbox, 1, (float)vpic.contrast);
	
	if(vcap.type&VID_TYPE_MONOCHROME)
	{
		slider_create("Whiteness", vbox, 2, (float)vpic.whiteness);
	}
	else
	{
		slider_create("Color", vbox, 3, (float)vpic.colour);
		slider_create("Hue", vbox, 4, (float)vpic.hue);
	}

	hbox = GTK_DIALOG(color_widget)->action_area;
	
	button = gtk_button_new_with_label("Dismiss");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)done_color_settings, NULL);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	gtk_widget_show(button);
	gtk_widget_grab_default(button);
	
	gtk_widget_show(hbox);
	gtk_widget_show(color_widget);
}

static void wide_toggle(void)
{
	int capt = capture_running;
	widescreen = 1 - widescreen;
	tv_capture(tv_fd, 0);
	tv_set_window(tv_fd);
	tv_capture(tv_fd, capt);
}

static void audio_toggle(void)
{
	set_audio(NULL, (void *)-1);
}

static void capture_toggle(void)
{
	if(!capture_running)
		tv_set_window(tv_fd);
	tv_capture(tv_fd, 1-capture_running);
	do_painting();
}

static int snapshot_chosen;
GtkWidget *file_selector;

static void snapshot_run(char *name)
{
	FILE *fp=fopen(name, "w");
	if(fp==NULL)
	{
		/* FIXME: write an alert dialog */
		fprintf(stderr,"Cannot open %s (%s)\n", 
			name, strerror(errno));
		return;
	}
	write_ppm_image(fp);
}

static void snapshot_clicked(GtkWidget *w, gpointer data)
{
	snapshot_chosen = (int)data;
	if(snapshot_chosen)
	{
		snapshot_run(gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector)));
	}
	gtk_widget_destroy(file_selector);
	file_selector=NULL;
}

static void snapshot_image(void)
{
	GtkWidget *filesel;
	
	if(file_selector)
		return;
	
	filesel = gtk_file_selection_new("Save as");
	
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
		"clicked", (GtkSignalFunc) snapshot_clicked, (void *)0);

	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
		"clicked", (GtkSignalFunc) snapshot_clicked, (void *)1);
		
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), "tv.pnm");
	gtk_widget_show(filesel);
	file_selector = filesel;
}
	
static void toolbar_add(GtkWidget *toolbar, char *name, void *func)
{	
	GtkWidget *button=gtk_button_new_with_label(name);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)func, NULL);
	gtk_box_pack_start(GTK_BOX(toolbar), button, TRUE, TRUE, 0);
	gtk_widget_show(button);
}

static void pack_toolbar(GtkWidget *toolbar)
{
	toolbar_add(toolbar, "Capture", capture_toggle);
	if(vcap.type&VID_TYPE_CAPTURE)
		toolbar_add(toolbar, "Snapshot", snapshot_image);
	toolbar_add(toolbar, "Picture", color_setting);
	if(vcap.audios)
		toolbar_add(toolbar, "Audio", audio_toggle);
	toolbar_add(toolbar, "Ch-", channel_down);
	toolbar_add(toolbar, "Ch+", channel_up);
	toolbar_add(toolbar, "Wide/Narr", wide_toggle);
	toolbar_add(toolbar, "Exit", gtk_main_quit);
}

void make_tv_set(void)
{
	GtkWidget *menubar;
	GtkWidget *hbox, *vbox;
	GtkWidget *toolbar;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
 
	gtk_widget_set_events(window, GDK_EXPOSURE_MASK|GDK_VISIBILITY_NOTIFY_MASK|GDK_ALL_EVENTS_MASK);
	gtk_signal_connect(GTK_OBJECT(window), "destroy", capture_off, NULL);
	gtk_signal_connect(GTK_OBJECT(window), "delete_event", gtk_main_quit, NULL);
	gtk_signal_connect(GTK_OBJECT(window), "destroy", capture_off, NULL);
	gtk_signal_connect(GTK_OBJECT(window), "delete_event", gtk_main_quit, NULL);
	gtk_signal_connect(GTK_OBJECT(window), "map_event", (GtkSignalFunc)
		painting_event, 0);
	gtk_signal_connect(GTK_OBJECT(window), "unmap_event", (GtkSignalFunc)
		painting_event, 0);
	gtk_signal_connect(GTK_OBJECT(window), "other_event", (GtkSignalFunc)
		painting_event, 0);
	gtk_signal_connect(GTK_OBJECT(window), "configure_event",
		(GtkSignalFunc)configure_event, NULL);


	vbox = gtk_vbox_new(FALSE, 0);

	toolbar = gtk_hbox_new(FALSE, 0);
	
	pack_toolbar(toolbar);
	
	gtk_widget_show(toolbar);
	
	/*
	 *	FIXME: now I can't resize below this user size. How do
	 *	I fix that ?
	 */
	 	
	hbox = tv = gtk_event_box_new();
	gtk_widget_set_usize(tv, xsize, ysize);
	if (pseudocolor)
	    make_color_cube();
	
	painter=gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(painter), xsize, ysize);
	gtk_widget_set_events(painter, GDK_EXPOSURE_MASK|GDK_VISIBILITY_NOTIFY_MASK|GDK_ALL_EVENTS_MASK);
	gtk_signal_connect(GTK_OBJECT(painter), "event", (GtkSignalFunc)
		painting_event, 0);
	gtk_signal_connect(GTK_OBJECT(painter), "map_event", (GtkSignalFunc)
		painting_event, 0);
	gtk_signal_connect(GTK_OBJECT(painter), "unmap_event", (GtkSignalFunc)
		painting_event, 0);
	gtk_signal_connect(GTK_OBJECT(painter), "other_event", (GtkSignalFunc)
		painting_event, 0);
	gtk_signal_connect(GTK_OBJECT(tv), "configure_event",
		(GtkSignalFunc)configure_event, NULL);
	gtk_widget_show(painter);
	gtk_container_add(GTK_CONTAINER(tv), painter);
	

	menubar = tv_menu_bar();
	gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), toolbar, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER(window), vbox);
	
	gtk_signal_connect(GTK_OBJECT(tv), "expose_event", 
		(GtkSignalFunc)painting_event, NULL);
	gtk_signal_connect(GTK_OBJECT(tv), "configure_event",
		(GtkSignalFunc)configure_event, NULL);
		
	gtk_widget_set_events(tv, GDK_STRUCTURE_MASK|GDK_EXPOSURE_MASK|GDK_ALL_EVENTS_MASK);

	gtk_widget_show(hbox);
	gtk_widget_show(vbox);
	gtk_widget_show(window);
	tv_set_window(tv_fd);
}

static GtkWidget *choose;
static int chosen_unit = 0;

static void video_device_ok(void)
{
	static void color_setting(void);
	if((tv_fd=open_tv_card(chosen_unit))==-1)
		return;
	gtk_widget_destroy(choose);
	make_tv_set();
	channel_up();	/* initialize tuner and title bar */
}

static void video_device_pick(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	chosen_unit = (int)data;
	
	if(event->type == GDK_2BUTTON_PRESS && event->button == 1)
		video_device_ok();
}

static void choose_video_device(void)
{
	GtkWidget *main_vbox, *vbox, *hbox, *button;
	GSList *owner;
	gint i;
	int count=0;
	
	choose = gtk_dialog_new();
	
	gtk_window_set_title(GTK_WINDOW(choose),"Select Video Device");
	main_vbox = GTK_DIALOG(choose)->vbox;
	
	vbox = gtk_vbox_new(FALSE,5);
	gtk_container_border_width(GTK_CONTAINER(vbox),3);
	gtk_box_pack_start(GTK_BOX(main_vbox), vbox, TRUE, TRUE, 0);
	gtk_widget_show(vbox);
	owner=NULL;
	
	for(i=0;i<240;i++)
	{
		char *p=probe_tv_set(i);
		if(p==NULL)
			continue;
		if(count==0)
		{
			chosen_unit=i;
		}
		count++;
		button=gtk_radio_button_new_with_label(owner, p);
		gtk_signal_connect(GTK_OBJECT(button), "button_press_event",
			(GtkSignalFunc)video_device_pick,
			(void *)i);
		gtk_box_pack_start(GTK_BOX(vbox), button, TRUE, TRUE, 0);
		gtk_widget_show(button);
		owner = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
	}
	hbox = GTK_DIALOG(choose)->action_area;
	
	button = gtk_button_new_with_label("OK");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)video_device_ok, NULL);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	gtk_widget_show(button);
	gtk_widget_grab_default(button);
	
	button = gtk_button_new_with_label("Cancel");
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)gtk_main_quit, NULL);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	gtk_widget_show(button);

#ifdef FIXME	
	if(count==0)
		video_no_devices();
#endif		
	if(count==1)
		video_device_ok();
	else
		gtk_widget_show(choose);
}




/*
 *	Run the TV set
 */

int main(int argc, char *argv[])
{
	int n=0;
	gtk_init(&argc,&argv);
	if(argv[1] && strcmp(argv[1], "-w")==0)
	{
		argv++;
		widescreen=1;
	}
	if(argv[1])
	{
		sscanf(argv[1],"%d",&n);
		if((tv_fd=open_tv_card(n))==-1)
			return 1;
		make_tv_set();
	}
	else
		choose_video_device();
	
	gtk_main();
	
	printf("Exiting\n");
	if(tv_fd != -1)
		close_tv_card(tv_fd);
	return 0;
}
